#include "generalChrNumModel.h"
#include "matrixUtils.h"
#include "someUtil.h"
#include "talRandom.h"



generalChrNumModel::generalChrNumModel(const finiteIntAlphabet* pAlph, const Vdouble& freq, rootFreqType freqType, vector<paramObj>& initParams)
: chrNumModel(pAlph, freq, freqType, IGNORE_PARAM) 
{
	init(initParams);
	updateQ();
	LOGDO(10,printMatrix(_Q, myLog::LogFile()));
}

generalChrNumModel::generalChrNumModel(const generalChrNumModel& other)
{
	copy(&other);
}

generalChrNumModel::~generalChrNumModel()
{
}

generalChrNumModel& generalChrNumModel::operator=(const generalChrNumModel&other)
{
	copy(&other);
	return *this;
}

void generalChrNumModel::copy(const generalChrNumModel* pOther)
{
	_gainConstR = pOther->_gainConstR;
	_gainLinearR = pOther->_gainLinearR;
	_lossConstR = pOther->_lossConstR;
	_lossLinearR = pOther->_lossLinearR;
	_duplConstR = pOther->_duplConstR;
	_duplLinearR = pOther->_duplLinearR;
	_validParams = pOther->_validParams;
	chrNumModel::copy(pOther);
}

void generalChrNumModel::init(const vector<paramObj>& initParams)
{
	_gainConstR = IGNORE_PARAM; 
	_gainLinearR = IGNORE_PARAM; 
	_lossConstR = IGNORE_PARAM; 
	_lossLinearR = IGNORE_PARAM; 
	_duplConstR = IGNORE_PARAM; 
	_duplLinearR = IGNORE_PARAM;
	_halfDuplR = IGNORE_PARAM;
	_scaleFactor = IGNORE_PARAM;
	bool bValidateParam = false;
	for (int i = 0; i < initParams.size(); ++i) 
	{
		if (!DEQUAL(initParams[i]._value,IGNORE_PARAM))
		{
			_validParams.push_back(initParams[i]);
			bValidateParam = true;
			switch (initParams[i]._type)
			{
				case LOSS_CONST:
					_lossConstR = initParams[i]._value;
					break;
				case LOSS_LINEAR:
					_lossLinearR = initParams[i]._value;
					break;
				case GAIN_CONST: 
					_gainConstR = initParams[i]._value;
					break;
				case GAIN_LINEAR: 
					_gainLinearR = initParams[i]._value;
					break;
				case DUPL: 
					_duplConstR = initParams[i]._value;
					break;
				case DUPL_LINEAR: 
					_duplLinearR = initParams[i]._value;
					break;
				case HALF_DUPL: 
					_halfDuplR = initParams[i]._value;
					break;
				case SCALE_BRANCH: 
					_scaleFactor = initParams[i]._value;
					break;
				default: 
					errorMsg::reportError("error in generalChrNumModel::init unknown param type");
					break;
			}
		}
	}
	if (!bValidateParam)
		errorMsg::reportError("error in generalChrNumModel::init there are no parameters to be included in the model");
	if ((_halfDuplR != IGNORE_PARAM) && (_duplLinearR != IGNORE_PARAM))
		errorMsg::reportError("not implemented yet: duplication rate linearly dependent on the current count with demiDuplication");
}

//reset the rate parameters to random values
void generalChrNumModel::setRandomRateParams(bool bUpdateQ)
{
	//note: _gainLinearR and _lossLinearR can be negative but larger than -_gainConst/maxChr so that all entries in the Q are positive
	//in such a case _gainConstR is also restricted such that the lower bound is max{0, g_l(1-maxNumber)} 
	if (!DEQUAL(_gainConstR, IGNORE_PARAM))
		_gainConstR = talRandom::giveRandomNumberBetweenTwoPoints(getParamLowerBound(GAIN_CONST), getParamUpperBound(GAIN_CONST));
	if (!DEQUAL(_gainLinearR, IGNORE_PARAM))
        _gainLinearR = talRandom::giveRandomNumberBetweenTwoPoints(getParamLowerBound(GAIN_LINEAR), getParamUpperBound(GAIN_LINEAR));
	if (!DEQUAL(_lossConstR, IGNORE_PARAM))
        _lossConstR = talRandom::giveRandomNumberBetweenTwoPoints(getParamLowerBound(LOSS_CONST), getParamUpperBound(LOSS_CONST));
	if (!DEQUAL(_lossLinearR, IGNORE_PARAM))
        _lossLinearR = talRandom::giveRandomNumberBetweenTwoPoints(getParamLowerBound(LOSS_LINEAR), getParamUpperBound(LOSS_LINEAR));
	if (!DEQUAL(_duplConstR, IGNORE_PARAM))
        _duplConstR = talRandom::giveRandomNumberBetweenTwoPoints(getParamLowerBound(DUPL), getParamUpperBound(DUPL));

	chrNumModel::setRandomRateParams(false);
	if (bUpdateQ == true)
		updateQ();
}

MDOUBLE generalChrNumModel::getParamUpperBound(paramType type) const
{
	MDOUBLE res = 0.0;
	switch (type)
	{
	case GAIN_LINEAR:
	case LOSS_LINEAR:
	case DUPL_LINEAR:
		res = upperValueOfLinearRateParam;
		break;
	case GAIN_CONST:
	case LOSS_CONST:
	case DUPL:
		res = upperValueOfRateParam;
		break;
	default:
		res = chrNumModel::getParamUpperBound(type);
	}
	return res;
}


MDOUBLE generalChrNumModel::getParamLowerBound(paramType type) const
{
	MDOUBLE res = 0.0;
	switch (type)
	{
	case GAIN_LINEAR:
		if (DEQUAL(_gainConstR, IGNORE_PARAM))
			res = lowerValueOfRateParam;
		else
			res = -_gainConstR / (_pAlph->max()-1);
		break;
	case LOSS_LINEAR:
		if (DEQUAL(_lossConstR, IGNORE_PARAM))
			res = lowerValueOfRateParam;
		else
			res = -_lossConstR / (_pAlph->max()-1);
		break;
	case DUPL_LINEAR:
		if (DEQUAL(_duplConstR, IGNORE_PARAM))
			res = lowerValueOfRateParam;
		else
			res = -_duplConstR / (_pAlph->max()-1);
		break;
	case LOSS_CONST:
		if (DEQUAL(_lossLinearR, IGNORE_PARAM))
            lowerValueOfRateParam;
		else
			res = (_lossLinearR > 0) ? lowerValueOfRateParam : (_lossLinearR * (1-_pAlph->max()));
		break;
	case GAIN_CONST:
		if (DEQUAL(_gainLinearR, IGNORE_PARAM))
            lowerValueOfRateParam;
		else
			res = (_gainLinearR > 0) ? lowerValueOfRateParam : (_gainLinearR * (1-_pAlph->max()));
        break;
	case DUPL:
		if (DEQUAL(_duplLinearR, IGNORE_PARAM))
            lowerValueOfRateParam;
		else
			res = (_duplLinearR > 0) ? lowerValueOfRateParam : (_duplLinearR * (1-_pAlph->max()));
        break;
	default:
		res = chrNumModel::getParamLowerBound(type);
	}
	return res;
}

bool generalChrNumModel::setGainConstR(MDOUBLE val) 
{ 
	if (val == _gainConstR)
		return false;
    _gainConstR = val; 
	updateQ();
	return true;
}


MDOUBLE generalChrNumModel::getGainR(int fromState) const
{
	MDOUBLE gainLinearR = DEQUAL(_gainLinearR, IGNORE_PARAM) ? 0.0 : _gainLinearR;
	MDOUBLE gainConstR = DEQUAL(_gainConstR, IGNORE_PARAM) ? 0.0 : _gainConstR;
	return gainConstR + fromState*gainLinearR;
}

MDOUBLE generalChrNumModel::getLossR(int fromState) const
{
    MDOUBLE lossLinearR = DEQUAL(_lossLinearR, IGNORE_PARAM) ? 0.0 : _lossLinearR;
	MDOUBLE lossConstR = DEQUAL(_lossConstR, IGNORE_PARAM) ? 0.0 : _lossConstR;
	return lossConstR + fromState*lossLinearR;
}

MDOUBLE generalChrNumModel::getDuplR(int fromState) const
{
 	MDOUBLE duplLinearR = DEQUAL(_duplLinearR, IGNORE_PARAM) ? 0.0 : _duplLinearR;
	MDOUBLE duplConstR = DEQUAL(_duplConstR, IGNORE_PARAM) ? 0.0 : _duplConstR;
	return duplConstR + fromState*duplLinearR;
}

MDOUBLE generalChrNumModel::getDemiR(int fromState) const
{
	if (DEQUAL(_halfDuplR, IGNORE_PARAM))
		return 0;
	else if (_halfDuplR == DEMI_EQUAL_DUPL)
		return getDuplR(fromState);
	else
        return _halfDuplR;
}


bool generalChrNumModel::setLossConstR(MDOUBLE val) 
{ 
	if (val == _lossConstR)
		return false;
    _lossConstR= val; 
	updateQ();
	return true;
}

bool generalChrNumModel::setDuplConstR(MDOUBLE val) 
{ 
	if (val == _duplConstR)
		return false;
    _duplConstR= val; 
	updateQ();
	return true;
}

bool generalChrNumModel::setDuplLinearR(MDOUBLE val)
{
	if (val == _duplLinearR)
		return false;
    _duplLinearR= val; 
	updateQ();
	return true;
}
bool generalChrNumModel::setLossLinearR(MDOUBLE val) 
{
	if (val == _lossLinearR)
		return false;
    _lossLinearR = val; 
	updateQ();
	return true;
}

bool generalChrNumModel::setGainLinearR(MDOUBLE val) 
{
	if (val == _gainLinearR)
		return false;
    _gainLinearR = val; 
	updateQ();
	return true;
}


void generalChrNumModel::getParams(vector<paramObj>& params) const
{
	params.clear();
	if (!DEQUAL(IGNORE_PARAM, _lossConstR)) {
		params.push_back(paramObj(LOSS_CONST, _lossConstR));
	}
	if (!DEQUAL(IGNORE_PARAM, _gainConstR)) {
		params.push_back(paramObj(GAIN_CONST, _gainConstR));
	}
	if (!DEQUAL(IGNORE_PARAM, _duplConstR)) {
		params.push_back(paramObj(DUPL, _duplConstR));
	}
	if (!DEQUAL(IGNORE_PARAM, _lossLinearR)) {
		params.push_back(paramObj(LOSS_LINEAR, _lossLinearR));
	}
	if (!DEQUAL(IGNORE_PARAM, _gainLinearR)) {
		params.push_back(paramObj(GAIN_LINEAR, _gainLinearR));
	}
	if (!DEQUAL(IGNORE_PARAM, _duplLinearR)) {
		params.push_back(paramObj(DUPL_LINEAR, _duplLinearR));
	}
	chrNumModel::getParams(params);
}


//void generalChrNumModel::updateQ()
//{
//	zeroMatrix(_Q);
//	//use local rates instead of asking if should IGNORE every time;
//    MDOUBLE lossLinearR = DEQUAL(_lossLinearR, IGNORE_PARAM) ? 0.0 : _lossLinearR;
//	MDOUBLE gainLinearR = DEQUAL(_gainLinearR, IGNORE_PARAM) ? 0.0 : _gainLinearR;
//	MDOUBLE duplLinearR = DEQUAL(_duplLinearR, IGNORE_PARAM) ? 0.0 : _duplLinearR;
//	MDOUBLE lossConstR = DEQUAL(_lossConstR, IGNORE_PARAM) ? 0.0 : _lossConstR;
//	MDOUBLE gainConstR = DEQUAL(_gainConstR, IGNORE_PARAM) ? 0.0 : _gainConstR;
//	MDOUBLE duplConstR = DEQUAL(_duplConstR, IGNORE_PARAM) ? 0.0 : _duplConstR;
//
//
//	//update first row (chrNum = 1): no loss possible. Q[1][2] = (g_c)+d
//	_Q[0][0] = - (gainConstR + duplConstR);
//	_Q[0][1] = gainConstR + duplConstR;
//
//    //update upper half of the matrix
//	int chr = 2;
//	MDOUBLE lossR, gainR, duplR;
//	for (; chr <= alphabetSize()/2; ++chr)
//	{
//		lossR = lossConstR + ((chr-1) * lossLinearR);
//		gainR = gainConstR + ((chr-1) * gainLinearR);
//		duplR = duplConstR + ((chr-1) * duplLinearR); 
//		_Q[chr-1][chr-2] = lossR;
//		_Q[chr-1][chr-1] = -lossR - gainR - duplR;
//		_Q[chr-1][chr] = gainR;
//		_Q[chr-1][2*chr -1] = duplR;
//	}
//	//update lower half of the matrix: duplication always go to last column
//	for (; chr < alphabetSize()-1; ++chr)
//	{
//		lossR = lossConstR + ((chr-1) * lossLinearR);
//		gainR = gainConstR + ((chr-1) * gainLinearR);
//		duplR = duplConstR + ((chr-1) * duplLinearR); 
//		_Q[chr-1][chr-2] = lossR;
//		_Q[chr-1][chr-1] = -lossR - gainR - duplR;
//		_Q[chr-1][chr] = gainR;
//		_Q[chr-1][alphabetSize()-1] = duplR;
//	}
//	//update one row before last: moving up is by gains+duplication 
//	lossR = lossConstR + ((chr-1) * lossLinearR);
//	gainR = gainConstR + ((chr-1) * gainLinearR);
//	duplR = duplConstR + ((chr-1) * duplLinearR);
//	_Q[chr-1][chr-2] = lossR;
//	_Q[chr-1][chr-1] = -lossR - gainR - duplR;
//	_Q[chr-1][chr] = gainR + duplR;
//
//	++chr;
//	lossR = lossConstR + ((chr-1) * lossLinearR);
//	gainR = gainConstR + ((chr-1) * gainLinearR);
//	duplR = duplConstR + ((chr-1) * duplLinearR); 
//
//	//update last row: no gains+duplication 
//	_Q[chr-1][chr-2] = lossR;
//	_Q[chr-1][chr-1] = -lossR;
//	if (_halfDuplR != IGNORE_PARAM)
//	{
//		if (_halfDuplR == DEMI_EQUAL_DUPL)
//            updateQHalfDupl(_duplConstR, _duplConstR);
//		else
//			updateQHalfDupl(_halfDuplR, _duplConstR);
//	}
//	_bQchanged = true;
//	_normQ = getMatrixNorm(_Q);
//}

void generalChrNumModel::updateQ()
{
	zeroMatrix(_Q);
	//use local rates instead of asking if should IGNORE every time;
    MDOUBLE lossLinearR = DEQUAL(_lossLinearR, IGNORE_PARAM) ? 0.0 : _lossLinearR;
	MDOUBLE gainLinearR = DEQUAL(_gainLinearR, IGNORE_PARAM) ? 0.0 : _gainLinearR;
	MDOUBLE duplLinearR = DEQUAL(_duplLinearR, IGNORE_PARAM) ? 0.0 : _duplLinearR;
	MDOUBLE lossConstR = DEQUAL(_lossConstR, IGNORE_PARAM) ? 0.0 : _lossConstR;
	MDOUBLE gainConstR = DEQUAL(_gainConstR, IGNORE_PARAM) ? 0.0 : _gainConstR;
	MDOUBLE duplConstR = DEQUAL(_duplConstR, IGNORE_PARAM) ? 0.0 : _duplConstR;
	
	int maxCount = _pAlph->max();
    //update upper half of the matrix
	MDOUBLE lossR, gainR, duplR;
	int id = 0;
	int count = _pAlph->id2Count(id);
	for (; count <= maxCount/2; ++id,count = _pAlph->id2Count(id))
	{
		if (id == 0)
		{//update first row (minimum count): no loss possible. Q[id][2id] = (g_c)+d
			_Q[id][id] = - (gainConstR + duplConstR);
			_Q[0][id+1] = gainConstR;
			_Q[0][_pAlph->count2Id(count*2)] += duplConstR;
			MDOUBLE x = _Q[0][_pAlph->count2Id(count*2)];
		}
		else 
		{
            lossR = lossConstR + ((count-1) * lossLinearR);
			gainR = gainConstR + ((count-1) * gainLinearR);
			duplR = duplConstR + ((count-1) * duplLinearR); 
			_Q[id][id-1] = lossR;
			_Q[id][id] = -lossR - gainR - duplR;
			_Q[id][id+1] = gainR;
			_Q[id][_pAlph->count2Id(count*2)] = duplR;
		}
	}
	//update lower half of the matrix: duplication always go to last column
	for (; count <= maxCount-2; ++id,count = _pAlph->id2Count(id))
	{
		_pAlph->id2Count(id);
		lossR = lossConstR + ((count-1) * lossLinearR);
		gainR = gainConstR + ((count-1) * gainLinearR);
		duplR = duplConstR + ((count-1) * duplLinearR); 
		_Q[id][id-1] = lossR;
		_Q[id][id] = -lossR - gainR - duplR;
		_Q[id][id+1] = gainR;
		_Q[id][alphabetSize()-1] = duplR;
	}
	//update one row before last: moving up is by gains+duplication 
	lossR = lossConstR + ((count-1) * lossLinearR);
	gainR = gainConstR + ((count-1) * gainLinearR);
	duplR = duplConstR + ((count-1) * duplLinearR);
	_Q[id][id-1] = lossR;
	_Q[id][id] = -lossR - gainR - duplR;
	_Q[id][id+1] = gainR + duplR;

	++id;
	count = _pAlph->id2Count(id);
	lossR = lossConstR + ((count-1) * lossLinearR);
	gainR = gainConstR + ((count-1) * gainLinearR);
	duplR = duplConstR + ((count-1) * duplLinearR); 

	//update last row: no gains+duplication 
	_Q[id][id-1] = lossR;
	_Q[id][id] = -lossR;
	if (_halfDuplR != IGNORE_PARAM)
	{
		if (_halfDuplR == DEMI_EQUAL_DUPL)
            updateQHalfDupl(_duplConstR, _duplConstR);
		else
			updateQHalfDupl(_halfDuplR, _duplConstR);
	}
	_bQchanged = true;
	_normQ = getMatrixNorm(_Q);
}

