#include "simulateChangesAlongTree.h"
#include "talRandom.h"
#include "matrixUtils.h"
#include "treeIt.h"
#include <algorithm>


simulateChangesAlongTree::simulateChangesAlongTree(const tree& inTree, const stochasticProcess& sp, alphabet* pAlph)
: _tree(inTree), _sp(sp), _pAlph(pAlph)	
{
}

simulateChangesAlongTree::~simulateChangesAlongTree()
{
}




void simulateChangesAlongTree::init()
{
	//init the vector of waiting times. 
	_waitingTimeParams.clear();
	_waitingTimeParams.resize(_pAlph->size());
	int i, j;
	for (i = 0; i < _pAlph->size(); ++i)
	{
		_waitingTimeParams[i] = -_sp.dPij_dt(i, i, 0.0);
		
	}

	//init _jumpProbs.
	//_jumpProbs[i][j] = Q[i][j] / -Q[i][i]
	_jumpProbs.clear();
	_jumpProbs.resize(_pAlph->size());
	for (i = 0; i < _pAlph->size(); ++i)
	{
		MDOUBLE sum = 0.0;
		_jumpProbs[i].resize(_pAlph->size());
		for (j = 0; j < _pAlph->size(); ++j)
		{
			if (i == j)
				_jumpProbs[i][j] = 0.0;
			else
			{
				_jumpProbs[i][j] = _sp.dPij_dt(i, j, 0.0) / _waitingTimeParams[i];
			}
			sum += _jumpProbs[i][j];
		}
		if (! DEQUAL(sum, 1.0)){
			string err = "error in simulateJumps::init(): sum probabilities is not 1 and equal to ";
			err+=double2string(sum);
			errorMsg::reportError(err);
		}
	}
	int nodesNum = _tree.getNodesNum();
	_changesOccurred.clear();
	_changesOccurred.resize(nodesNum);
	for (int i=0; i<nodesNum; ++i)
		resizeMatrix(_changesOccurred[i], _pAlph->size(), _pAlph->size());
	_nodesContent.clear();
	_nodesContent.resize(nodesNum, 0);
}

sequenceContainer simulateChangesAlongTree::simulatePosition(){
	init();
	Vdouble freqs(_pAlph->size(),0.0);
	for (int i = 0; i< freqs.size(); ++i)
		freqs[i]=_sp.freq(i);
	int rootState = giveRandomState(_pAlph->size(), freqs);
	string name = _tree.getRoot()->name();
	sequence seq(_pAlph->fromInt(rootState),name, "", _tree.getRoot()->id(), _pAlph);
	_sc.add(seq);



	_nodesContent[_tree.getRoot()->id()] = rootState;
	simulateOnce(_tree.getRoot(),0,rootState,0);
	simulateOnce(_tree.getRoot(),0,rootState,1);
	if (_tree.getRoot()->getNumberOfSons() > 2)
		simulateOnce(_tree.getRoot(),0,rootState,2);
	calcEventsTypes();
	return _sc;
}

void simulateChangesAlongTree::simulateOnce(tree::nodeP curNode, 
											MDOUBLE disFromNode, 
											int previousContent, int whichSon){
		tree::nodeP sonNode = curNode->getSon(whichSon);
		MDOUBLE avgWaitingTime = 1.0 / _waitingTimeParams[previousContent];
		MDOUBLE timeTillChange = talRandom::rand_exp(avgWaitingTime);
		disFromNode += timeTillChange;
		int nextContent = giveRandomState(_pAlph->size(), previousContent, _jumpProbs);
		while (disFromNode < sonNode->dis2father()) {
			_changesOccurred[sonNode->id()][previousContent][nextContent]++;
			previousContent=nextContent;
			MDOUBLE avgWaitingTime = 1.0 / _waitingTimeParams[previousContent];
			MDOUBLE timeTillChange = talRandom::rand_exp(avgWaitingTime);
			disFromNode += timeTillChange;
			nextContent = giveRandomState(_pAlph->size(), nextContent, _jumpProbs);
		}
		while (disFromNode >= sonNode->dis2father()) {
			_nodesContent[sonNode->id()] = previousContent;
			string name = sonNode->name();
			sequence seq(_pAlph->fromInt(previousContent),name, "", sonNode->id(), _pAlph);
			_sc.add(seq);
			if (sonNode->isLeaf()) {
				//string name = sonNode->name();
				//sequence seq(int2string(previousContent),name, "", sonNode->id(), _pAlph);
				//_sc.add(seq);
				return;
			}
			for (int s = 1; s < sonNode->getNumberOfSons(); ++s)
                simulateOnce(sonNode, 0, previousContent, s);
			disFromNode-=sonNode->dis2father();
			curNode = sonNode;
			sonNode = curNode->getSon(0);
		}
		_changesOccurred[sonNode->id()][previousContent][nextContent]++;
		simulateOnce(curNode, disFromNode, nextContent, 0);
}

VVint simulateChangesAlongTree::getChangesForBranch(int nodeID){
	if (nodeID>_changesOccurred.size())
		errorMsg::reportError("error in simulateChangesAlongTree::getChangesForBranch, nodeID doesn't exist");
	return _changesOccurred[nodeID];
}

sequenceContainer simulateChangesAlongTree::toSeqDataWithoutInternalNodes() {
	sequenceContainer myseqData;
	int nextId = 0;
	for (sequenceContainer::taxaIterator it =_sc.taxaBegin(); it != _sc.taxaEnd(); ++it)
	{
		string nodeName = it->name();
		tree::nodeP theCurNode = _tree.findNodeByName(nodeName);
		if (theCurNode == NULL)
			errorMsg::reportError("could not find the specified name: " + nodeName);
		if (theCurNode->isInternal()) 
			continue;
		sequence tmpSeq = *it;
		tmpSeq.setID(nextId);
		myseqData.add(tmpSeq);
		++nextId;
	}
	return myseqData;
}

//type = gain/loss/dupl/demi/ or jumps to maxChr
int simulateChangesAlongTree::getTotalJumps(chrNumModel::jumpType type)
{
	int totalPos = _changesTypes.size() -1;
	switch (type)
	{
	case chrNumModel::GAIN_J:
	case chrNumModel::LOSS_J:
	case chrNumModel::DUPL_J:
	case chrNumModel::DEMI_J:
	case chrNumModel::MAX_CHR_J:
		return _changesTypes[totalPos][type];
		break;
	default:
		errorMsg::reportError("simulateChangesAlongTree::getTotalJumps: unknown type");
		return -1;
	}
}

//fill _changesTypes: the number of gain/loss/dupl/demi/to_max_chr for each node
void simulateChangesAlongTree::calcEventsTypes()
{
	int maxChr = _pAlph->size();
	int nodesNum = _tree.getNodesNum();
	_changesTypes.resize(nodesNum+1);
	for (int n = 0; n < _changesTypes.size(); ++n)
		_changesTypes[n].resize(chrNumModel::JUMP_TYPE_MAX, 0);
	treeIterTopDownConst tIt(_tree);
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
	{
		if (mynode->isRoot())
			continue;
		int nodeId = mynode->id();
		int chr;
		for (chr = 2; chr < maxChr-1; ++chr)
		{
			_changesTypes[nodeId][chrNumModel::GAIN_J] += _changesOccurred[nodeId][chr-1][chr];
			_changesTypes[nodeId][chrNumModel::LOSS_J] += _changesOccurred[nodeId][chr-1][chr-2];
			_changesTypes[nodeId][chrNumModel::MAX_CHR_J] += _changesOccurred[nodeId][chr-1][maxChr-1]; //changes to maxChr

			//compute exp of a duplication 
			if (chr <= maxChr/2) 
			{
				_changesTypes[nodeId][chrNumModel::DUPL_J] += _changesOccurred[nodeId][chr-1][2*chr-1];
			}
			//else 
			//{
			//	_changesTypes[nodeId][2] += _changesOccurred[nodeId][chr-1][maxChr-1];
			//}
			//compute exp of a demi-duplication 
			if ((chr <= maxChr * 0.66) && (chr > 2))
			{
				if (chr == 3)
				{
					_changesTypes[nodeId][chrNumModel::DEMI_J] += _changesOccurred[nodeId][2][4]; //count only 3->5
				}
				else if ((chr % 2) == 0)
				{
                    _changesTypes[nodeId][chrNumModel::DEMI_J] += _changesOccurred[nodeId][chr-1][static_cast<int>(1.5*chr-1)];
				}
				else
				{
					_changesTypes[nodeId][chrNumModel::DEMI_J] += _changesOccurred[nodeId][chr-1][static_cast<int>(1.5*chr-1)];
					_changesTypes[nodeId][chrNumModel::DEMI_J] += _changesOccurred[nodeId][chr-1][static_cast<int>(1.5*chr)];
				}
			}
		}
		//for (maxChr-1) there are only two possibilities : loss and jumps to maxChr
		chr = maxChr-1;
		_changesTypes[nodeId][chrNumModel::LOSS_J] += _changesOccurred[nodeId][chr-1][chr-2];
		_changesTypes[nodeId][chrNumModel::MAX_CHR_J] += _changesOccurred[nodeId][chr-1][maxChr-1];
		//for maxChr there is only one possibility loss 
		chr = maxChr;
        _changesTypes[nodeId][chrNumModel::LOSS_J] += _changesOccurred[nodeId][chr-1][chr-2]; //loss from max to max-1
		//add all results
		_changesTypes[nodesNum][chrNumModel::GAIN_J] += _changesTypes[nodeId][chrNumModel::GAIN_J];
		_changesTypes[nodesNum][chrNumModel::LOSS_J] += _changesTypes[nodeId][chrNumModel::LOSS_J];
		_changesTypes[nodesNum][chrNumModel::DUPL_J] += _changesTypes[nodeId][chrNumModel::DUPL_J];
		_changesTypes[nodesNum][chrNumModel::DEMI_J] += _changesTypes[nodeId][chrNumModel::DEMI_J];
		_changesTypes[nodesNum][chrNumModel::MAX_CHR_J] += _changesTypes[nodeId][chrNumModel::MAX_CHR_J];
	}	
}

void simulateChangesAlongTree::printEvents(ostream &expFile)
{
	treeIterTopDownConst tItr(_tree);
	//print gains
	expFile<<"#Nodes with GAIN events"<<endl;
	MDOUBLE sum = 0.0;
	for (tree::nodeP mynode = tItr.first(); mynode != tItr.end(); mynode = tItr.next()) 
	{
		if (mynode->isRoot())
			continue;
		int nodeId = mynode->id();
		sum += _changesTypes[nodeId][chrNumModel::GAIN_J];
		if (_changesTypes[nodeId][chrNumModel::GAIN_J] > 0)
			expFile<<mynode->name()<<": "<<_changesTypes[nodeId][chrNumModel::GAIN_J]<<endl;
	}
	expFile<<"Total number of gain events: "<<sum<<endl;

	//print losses
	expFile<<"#+++++++++++++++++++++++++++++"<<endl<<endl;
	expFile<<"#Nodes with LOSS events"<<endl;
	sum = 0.0;
	for (tree::nodeP mynode = tItr.first(); mynode != tItr.end(); mynode = tItr.next()) 
	{
		if (mynode->isRoot())
			continue;
		int nodeId = mynode->id();
		sum += _changesTypes[nodeId][chrNumModel::LOSS_J];
		if (_changesTypes[nodeId][chrNumModel::LOSS_J] > 0)
			expFile<<mynode->name()<<": "<<_changesTypes[nodeId][chrNumModel::LOSS_J]<<endl;
	}
	expFile<<"Total number of loss events: "<<sum<<endl;

	//print duplications
	expFile<<"#+++++++++++++++++++++++++++++"<<endl<<endl;
	expFile<<"#Nodes with duplication events"<<endl;
	sum = 0.0;
	for (tree::nodeP mynode = tItr.first(); mynode != tItr.end(); mynode = tItr.next()) 
	{
		if (mynode->isRoot())
			continue;
		int nodeId = mynode->id();
		sum += _changesTypes[nodeId][chrNumModel::DUPL_J];
		if (_changesTypes[nodeId][chrNumModel::DUPL_J] > 0)
			expFile<<mynode->name()<<": "<<_changesTypes[nodeId][chrNumModel::DUPL_J]<<endl;
	}
	expFile<<"Total number of duplication events: "<<sum<<endl;

	//print demi-duplications
	expFile<<"#+++++++++++++++++++++++++++++"<<endl<<endl;
	expFile<<"#Nodes with demi-dulications events"<<endl;
	sum = 0.0;
	for (tree::nodeP mynode = tItr.first(); mynode != tItr.end(); mynode = tItr.next()) 
	{
		if (mynode->isRoot())
			continue;
		int nodeId = mynode->id();
		sum += _changesTypes[nodeId][chrNumModel::DEMI_J];
		if (_changesTypes[nodeId][chrNumModel::DEMI_J] > 0)
			expFile<<mynode->name()<<": "<<_changesTypes[nodeId][chrNumModel::DEMI_J]<<endl;
	}
	expFile<<"Total number of demi-dulications events: "<<sum<<endl;
	expFile<<"#+++++++++++++++++++++++++++++"<<endl<<endl;

	//print transitions to max chromosome
	expFile<<"#+++++++++++++++++++++++++++++"<<endl<<endl;
	expFile<<"#Nodes with transitions to max chromosome allowed"<<endl;
	sum = 0.0;
	for (tree::nodeP mynode = tItr.first(); mynode != tItr.end(); mynode = tItr.next()) 
	{
		if (mynode->isRoot())
			continue;
		int nodeId = mynode->id();
		sum += _changesTypes[nodeId][chrNumModel::MAX_CHR_J];
		if (_changesTypes[nodeId][chrNumModel::MAX_CHR_J] > 0)
			expFile<<mynode->name()<<": "<<_changesTypes[nodeId][chrNumModel::MAX_CHR_J]<<endl;
	}
	expFile<<"Total number of transitions to max chromosome: "<<sum<<endl;
}
